import boto3
import requests
from os.path import expanduser, join, exists
from configparser import ConfigParser
from .config import format_agent_profile_name
from configparser import ConfigParser
import time
import boto3
from os import mkdir
import time
from .exceptions import ConfigProfileNotSet, PennsieveAccountInformationFailedAuthentication

from ..constants import PENNSIEVE_URL

from .logger import logger

from .profile import create_unique_profile_name

userpath = expanduser("~")
configpath = join(userpath, ".pennsieve", "config.ini")

# Variables for token caching

cached_access_token = None
last_fetch_time = 0
TOKEN_CACHE_DURATION = 60 # Amount of time in seconds to cache the access token




def get_access_token(api_key=None, api_secret=None):
    """
        Creates a temporary access token for utilizing the Pennsieve API. Reads the api token and secret from the Pennsieve config.ini file.
        get cognito config . If no target profile name is provided the default profile is used. 
    """
    global cached_access_token, last_fetch_time, TOKEN_CACHE_DURATION # Variables used for token caching
    global logger
    current_time = time.time()

    # If the cached_access_token is not None and the last fetch time is less than the cache duration, return the cached access token
    if cached_access_token and current_time - last_fetch_time < TOKEN_CACHE_DURATION:
        return cached_access_token
    
    r = requests.get(f"{PENNSIEVE_URL}/authentication/cognito-config")
    r.raise_for_status()

    cognito_app_client_id = r.json()["tokenPool"]["appClientId"]
    cognito_region_name = r.json()["region"]

    cognito_idp_client = boto3.client(
        "cognito-idp",
        region_name=cognito_region_name,
        aws_access_key_id="",
        aws_secret_access_key="",
    )

    # use the default profile values for auth if no api_key or api_secret is provided
    if api_key is None or api_secret is None:
        api_key = get_profile_name_from_api_key("api_token")
        api_secret = get_profile_name_from_api_key("api_secret")




    login_response = cognito_idp_client.initiate_auth(
    AuthFlow="USER_PASSWORD_AUTH",
    AuthParameters={"USERNAME": api_key, "PASSWORD": api_secret},
    ClientId=cognito_app_client_id,
    )

    cached_access_token = login_response["AuthenticationResult"]["AccessToken"]
    last_fetch_time = current_time

    return cached_access_token
        

def clear_cached_access_token():
    global cached_access_token, last_fetch_time
    cached_access_token = None


def get_cognito_userpool_access_token(email, password):
    """
    Creates a temporary access token for utilizing the Pennsieve API. Utilizes email and password to authenticate with the Pennsieve Cognito Userpool 
    which provides higher privileges than the API token and secret flow.
    """
    

    try:
        response = requests.get(f"{PENNSIEVE_URL}/authentication/cognito-config")
        response.raise_for_status()
        cognito_app_client_id = response.json()["userPool"]["appClientId"]
        cognito_region = response.json()["userPool"]["region"]
        cognito_client = boto3.client(
            "cognito-idp",
            region_name=cognito_region,
            aws_access_key_id="",
            aws_secret_access_key="",
        )
    except Exception as e:
        raise Exception(e) from e

    try:
        login_response = cognito_client.initiate_auth(
            AuthFlow="USER_PASSWORD_AUTH",
            AuthParameters={"USERNAME": email, "PASSWORD": password},
            ClientId=cognito_app_client_id,
        )
    except Exception as e:
        raise PennsieveAccountInformationFailedAuthentication("Invalid email or password") from e
    try:
        access_token = login_response["AuthenticationResult"]["AccessToken"]
        response = requests.get(
            f"{PENNSIEVE_URL}/user", headers={"Authorization": f"Bearer {access_token}"}
        )
        response.raise_for_status()
    except Exception as e:
        raise e

    return access_token

def get_profile_name_from_api_key(key):
    config = ConfigParser()
    config.read(configpath)
    if "global" not in config:
        raise Exception("Profile has not been set")

    keyname = config["global"]["default_profile"]

    if keyname in config and key in config[keyname]:
        return config[keyname][key]
    return None


def bf_delete_account(keyname):
    """
    Args:
        keyname: name of local Pennsieve account key (string)
    Action:
        Deletes account information from the Pennsieve config file
    """
    config = ConfigParser()
    config.read(configpath)
    config.remove_section(keyname)
    with open(configpath, "w") as configfile:
        config.write(configfile)

    

def bf_delete_default_profile():
    config = ConfigParser()
    config.read(configpath)

    if "global" not in config:
        return 
    
    config.remove_section("global")

    with open(configpath, "w") as configfile:
        config.write(configfile)




def bf_add_account_username(keyname, key, secret):
    """
    Associated with 'Add account' button in 'Login to your Pennsieve account' section of SODA

    Args:
        keyname: Name of the account to be associated with the given credentials (string)
        key: API key (string)
        secret: API Secret (string)
    Action:
        Adds account to the Pennsieve configuration file (local machine)
    """
    global logger


    # format the keyname to lowercase and replace '.' with '_'
    formatted_account_name = format_agent_profile_name(keyname)
    
    bf_delete_default_profile()
    try:
        bfpath = join(userpath, ".pennsieve")
        # Load existing or create new config file
        config = ConfigParser()
        if exists(configpath):
            config.read(configpath)
        elif not exists(bfpath):
            mkdir(bfpath)

        # Add agent section
        agentkey = "agent"
        if not config.has_section(agentkey):
            config.add_section(agentkey)
            config.set(agentkey, "port", "9000")
            config.set(agentkey, "upload_workers", "10")
            config.set(agentkey, "upload_chunk_size", "32")

        # Add new account if it does not already exist
        if not config.has_section(formatted_account_name):
            config.add_section(formatted_account_name)

        config.set(formatted_account_name, "api_token", key)
        config.set(formatted_account_name, "api_secret", secret)

        # set profile name in global section
        if not config.has_section("global"):
            config.add_section("global")
            config.set("global", "default_profile", formatted_account_name)

        
        with open(configpath, "w") as configfile:
            config.write(configfile)

    except Exception as e:
        raise e

    # Check key and secret are valid, if not delete account from config
    try:
        get_access_token()
    except Exception as e:
        logger.error(e)
        bf_delete_account(formatted_account_name)
        raise PennsieveAccountInformationFailedAuthentication("Pkease check that the key name, key value, and secret value are entered properly.") from e

    try:
        if not config.has_section("global"):
            config.add_section("global")

        default_acc = config["global"]
        default_acc["default_profile"] = keyname

        with open(configpath, "w+") as configfile:
            config.write(configfile)

        return {"message": f"Successfully added account {formatted_account_name}"}

    except Exception as e:
        bf_delete_account(keyname)
        raise e


def delete_duplicate_keys(token, keyname):
    try:

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {token}",
        }

        r = requests.get(f"{PENNSIEVE_URL}/token", headers=headers)
        r.raise_for_status()

        tokens = r.json()

        for token in tokens:
            if token["name"] == keyname:
                r = requests.delete(f"{PENNSIEVE_URL}/token/{token['key']}", headers=headers)
                r.raise_for_status()
    except Exception as e:
        raise e


def create_pennsieve_api_key_secret(email, password, machine_username_specifier):

    api_key = get_cognito_userpool_access_token(email, password)

    # TODO: Send in computer and profile of computer from frontend to this endpoint and use it in this function
    profile_name = create_unique_profile_name(api_key, machine_username_specifier)


    delete_duplicate_keys(api_key, "SODA-Pennsieve")
    delete_duplicate_keys(api_key, profile_name)


    url = "https://api.pennsieve.io/token/"

    payload = {"name": f"{profile_name}"}
    headers = {
        "Accept": "*/*",
        "Content-Type": "application/json",
        "Authorization": f"Bearer {api_key}",
    }

    response = requests.request("POST", url, json=payload, headers=headers)
    response.raise_for_status()
    response = response.json()


    # clear access token cache
    clear_cached_access_token()

    return { 
        "success": "success", 
        "key": response["key"], 
        "secret": response["secret"], 
        "name": profile_name
    }


def get_access_token(api_key=None, api_secret=None):
    """
        Creates a temporary access token for utilizing the Pennsieve API. Reads the api token and secret from the Pennsieve config.ini file.
        get cognito config . If no target profile name is provided the default profile is used. 
    """
    global cached_access_token, last_fetch_time, TOKEN_CACHE_DURATION # Variables used for token caching
    # global logger
    current_time = time.time()

    print("Current time:", current_time)



    # If the cached_access_token is not None and the last fetch time is less than the cache duration, return the cached access token
    if cached_access_token and current_time - last_fetch_time < TOKEN_CACHE_DURATION:
        return cached_access_token
    
    print("Cached token not returned")
    r = requests.get(f"{PENNSIEVE_URL}/authentication/cognito-config")
    r.raise_for_status()
    print("Cognito config response:", r.json())

    cognito_app_client_id = r.json()["tokenPool"]["appClientId"]
    cognito_region_name = r.json()["region"]

    cognito_idp_client = boto3.client(
        "cognito-idp",
        region_name=cognito_region_name,
        aws_access_key_id="",
        aws_secret_access_key="",
    )

    print("boto3 client done")

    # use the default profile values for auth if no api_key or api_secret is provided
    if api_key is None or api_secret is None:
        api_key = get_profile_name_from_api_key("api_token")
        api_secret = get_profile_name_from_api_key("api_secret")


    login_response = cognito_idp_client.initiate_auth(
    AuthFlow="USER_PASSWORD_AUTH",
    AuthParameters={"USERNAME": api_key, "PASSWORD": api_secret},
    ClientId=cognito_app_client_id,
    )

    cached_access_token = login_response["AuthenticationResult"]["AccessToken"]
    last_fetch_time = current_time

    return cached_access_token

def get_profile_name_from_api_key(key):
    config = ConfigParser()
    config.read(configpath)
    if "global" not in config:
        raise ConfigProfileNotSet("global")

    keyname = config["global"]["default_profile"]

    if keyname in config and key in config[keyname]:
        return config[keyname][key]
    return None


def create_request_headers(ps_or_token):
    """
    Creates necessary HTTP headers for making Pennsieve API requests.
    Input: 
        ps: Pennsieve object for a user that has been authenticated
    """
    if type(ps_or_token) == str:
        return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {ps_or_token}",
        }
    
    return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {ps_or_token.get_user().session_token}",
    }


import requests
from ..constants import PENNSIEVE_URL
from .authentication import get_access_token
import re 
from .exceptions import PennsieveDatasetCannotBeFound, FailedToFetchPennsieveDatasets

def get_dataset_id(dataset_name_or_id):
    """
    Returns the dataset ID for the given dataset name.
    If the dataset ID was provided instead of the name, the ID will be returned. *Common for Guided Mode*
    
    Input:
        dataset_name_or_id: Pennsieve dataset name or ID to get the ID for
    """
    # If the input is already a dataset ID, return it
    if dataset_name_or_id.startswith("N:dataset:"):
        return dataset_name_or_id
    

    try:
        # Attempt to retrieve the user's dataset list from Pennsieve
        dataset_list = get_users_dataset_list()
    except Exception as e:
        raise FailedToFetchPennsieveDatasets(str(e))
    
    # Iterate through the user's dataset list to find a matching dataset name
    for dataset in dataset_list:
        if dataset["content"]["name"] == dataset_name_or_id:
            return dataset["content"]["id"]
    
    # If no matching dataset is found, abort with a 404 status and a specific error message
    raise PennsieveDatasetCannotBeFound(dataset_name_or_id)
  

def get_users_dataset_list():
    """
        Returns a list of datasets the user has access to.
        Input:
            token: Pennsieve access token
    """

    # The number of datasets to retrieve per chunk
    NUMBER_OF_DATASETS_PER_CHUNK = 200
    # The total number of datasets the user has access to (set after the first request)
    NUMBER_OF_DATASETS_USER_HAS_ACCESS_TO = None

    # The offset is the number of datasets to skip before retrieving the next chunk of datasets (starts at 0, then increases by the number of datasets per chunk)
    current_offset = 0
    # The list of datasets the user has access to (datasets are added to this list after each request and then returned)
    datasets = []

    try:
        # Get the first chunk of datasets as well as the total number of datasets the user has access to
        r = requests.get(f"{PENNSIEVE_URL}/datasets/paginated", headers=create_request_headers(get_access_token()), params={"offset": current_offset, "limit": NUMBER_OF_DATASETS_PER_CHUNK})
        r.raise_for_status()
        responseJSON = r.json()
        datasets.extend(responseJSON["datasets"])
        NUMBER_OF_DATASETS_USER_HAS_ACCESS_TO = responseJSON["totalCount"]

        # If the number of datasets the user has access to is less than the number of datasets per chunk, we don't need to retrieve any more datasets
        if NUMBER_OF_DATASETS_USER_HAS_ACCESS_TO < NUMBER_OF_DATASETS_PER_CHUNK:
            return datasets
        
        # Otherwise, we need to retrieve the rest of the datasets.
        # We do this by retrieving chunks of datasets until the number of datasets retrieved is equal to the number of datasets the user has access to
        while len(datasets) < NUMBER_OF_DATASETS_USER_HAS_ACCESS_TO:
            # Increase the offset by the number of datasets per chunk (e.g. if 200 datasets per chunk, then increase the offset by 200)
            current_offset += NUMBER_OF_DATASETS_PER_CHUNK
            r = requests.get(f"{PENNSIEVE_URL}/datasets/paginated", headers=create_request_headers(get_access_token()), params={"offset": current_offset, "limit": NUMBER_OF_DATASETS_PER_CHUNK})
            r.raise_for_status()
            responseJSON = r.json()
            datasets.extend(responseJSON["datasets"])

        return datasets
    except Exception as e:
        raise e
    





def create_request_headers(ps_or_token):
    """
    Creates necessary HTTP headers for making Pennsieve API requests.
    Input: 
        ps: Pennsieve object for a user that has been authenticated
    """
    if type(ps_or_token) == str:
        return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {ps_or_token}",
        }
    
    return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {ps_or_token.get_user().session_token}",
    }


forbidden_characters_bf = '\/:*?"<>.,'


def check_forbidden_characters_ps(my_string):
    """
    Check for forbidden characters in Pennsieve file/folder name

    Args:
        my_string: string with characters (string)
    Returns:
        False: no forbidden character
        True: presence of forbidden character(s)
    """
    regex = re.compile(f"[{forbidden_characters_bf}]")
    if regex.search(my_string) == None and "\\" not in r"%r" % my_string:
        return False
    else:
        return True


import logging

from .permissions import has_edit_permissions



# Create a logger for the package
logger = logging.getLogger(__name__)
# Optional: Provide a default configuration if no handlers are set
logger.setLevel(logging.WARNING)


from os.path import join, getsize, abspath, dirname, expanduser
from os import makedirs



TEMPLATE_PATH = join(dirname(abspath(__file__)), '..', 'metadata_templates')
METADATA_UPLOAD_PS_PATH = expanduser("~/.pysoda")
makedirs(METADATA_UPLOAD_PS_PATH, exist_ok=True)


SCHEMA_NAMES = {
    "submission": "submission_schema.json",
    "subjects": "subjects_schema.json"
}


SDS_FILE_SUBJECTS = "subjects.xlsx"
SCHEMA_NAME_SUBJECTS = "subjects.json"
SDS_FILE_SAMPLES = "samples.xlsx"
SCHEMA_NAME_SAMPLES = "samples.json"
SDS_FILE_PERFORMANCES = "performances.xlsx"
SCHEMA_NAME_PERFORMANCES = "performances.json"
SDS_FILE_SITES = "sites.xlsx"
SCHEMA_NAME_SITES = "sites.json"
SDS_FILE_RESOURCES = "resources.xlsx"
SCHEMA_NAME_RESOURCES = "resources.json"
SDS_FILE_DATASET_DESCRIPTION = "dataset_description.xlsx"
SCHEMA_NAME_DATASET_DESCRIPTION = "dataset_description.json"
SDS_FILE_CODE_DESCRIPTION = "code_description.xlsx"
SCHEMA_NAME_CODE_DESCRIPTION = "code_description.json"
SDS_FILE_MANIFEST = "manifest.xlsx"
SCHEMA_NAME_MANIFEST = "manifest.json"


from .upload_utils import generate_options_set, generating_locally, generating_on_ps, uploading_with_ps_account, uploading_to_existing_ps_dataset, can_resume_prior_upload, virtual_dataset_empty, get_dataset_id
from .exceptions import PropertyNotSetError, ConfigProfileNotSet, FailedToFetchPennsieveDatasets, FailedToFetchPennsieveDatasets, PennsieveActionNoPermission, PennsieveDatasetCannotBeFound, EmptyDatasetError, LocalDatasetMissingSpecifiedFiles, validation_error_message, GenericUploadError, PennsieveUploadException, PennsieveDatasetNameTaken, PennsieveDatasetNameInvalid, PennsieveAccountInvalid, GenerateOptionsNotSet, PennsieveDatasetFilesInvalid, PennsieveAgentError, PennsieveAccountInformationFailedAuthentication
from .pennsieveAgentUtils import connect_pennsieve_client
from .schema_validation import validate_schema, get_sds_headers, get_schema_path
from .config import format_agent_profile_name
from .pennsieveUtils import get_dataset_id, check_forbidden_characters_ps, get_users_dataset_list
from .authentication import get_access_token, create_request_headers
from .metadata_utils import column_check, returnFileURL, remove_high_level_folder_from_path, double_extensions, get_name_extension
from .time_utils import TZLOCAL



# create a custom exception that indicates a property in an object has not been set
class PropertyNotSetError(Exception):
    def __init__(self, property_name):
        self.property_name = property_name
        self.error_message = f"The property {self.property_name} has not been set."

    def __str__(self):
        return self.error_message
    

# create a custom exception that indicates that the 'PennsieveAgent' could not be started
class PennsieveAgentError(Exception):
    def __init__(self, error_message):
        self.error_message = error_message

    def __str__(self):
        return self.error_message
    
class FailedToFetchPennsieveDatasets(Exception):
    def __init__(self, error_message):
        self.error_message = error_message

    def __str__(self):
        return self.error_message
    
class PennsieveDatasetCannotBeFound(Exception):
    def __init__(self, dataset_name):
        self.dataset_name = dataset_name
        self.error_message = f"The Pennsieve dataset {self.dataset_name} could not be found."

    def __str__(self):
        return self.error_message
    
    
class ConfigProfileNotSet(Exception):
    def __init__(self, profile_name):
        self.profile_name = profile_name
        self.error_message = f"The profile {self.profile_name} has not been set."

    def __str__(self):
        return self.error_message
    
class GenerateOptionsNotSet(Exception):
    def __init__(self):
        self.error_message = "The generate options have not been set."

    def __str__(self):
        return self.error_message
    

class PennsieveActionNoPermission(Exception):
    def __init__(self, action):
        self.action = action
        self.error_message = f"Do not have the correct permissions to perform action: {self.action} ."

    def __str__(self):
        return self.error_message

class GenericUploadError(Exception):
    def __init__(self, error_message):
        self.error_message = error_message

    def __str__(self):
        return self.error_message
    

class EmptyDatasetError(Exception):
    def __init__(self, dataset_name, expanded=""):
        self.dataset_name = dataset_name
        self.expanded = expanded
        self.error_message = f"The dataset {self.dataset_name} is empty. {expanded}"

    def __str__(self):
        return self.error_message
    

class LocalDatasetMissingSpecifiedFiles(Exception):
    def __init__(self, error_message):
        self.error_message = error_message

    def __str__(self):
        return self.error_message


class PennsieveUploadException(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.error_message


class PennsieveAccountInformationFailedAuthentication(Exception):
    def __init__(self, error_message):

        self.error_message = error_message

    def __str__(self):
        return self.error_message


class PennsieveDatasetNameTaken(Exception):
    def __init__(self, dataset_name):
        self.dataset_name = dataset_name
        self.error_message = f"The Pennsieve dataset name {self.dataset_name} is already taken."

    def __str__(self):
        return self.error_message
    
class PennsieveDatasetNameInvalid(Exception):
    def __init__(self, dataset_name):
        self.dataset_name = dataset_name
        self.error_message = f"The Pennsieve dataset name {self.dataset_name} is invalid."

    def __str__(self):
        return self.error_message
    
class PennsieveAccountInvalid(Exception):
    def __init__(self, account_name):
        self.account_name = account_name
        self.error_message = f"The Pennsieve account name {self.account_name} is invalid."

    def __str__(self):
        return self.error_message
    
class PennsieveDatasetFilesInvalid(Exception):
    def __init__(self, error_message):
        self.error_message = error_message

    def __str__(self):
        return self.error_message

def validation_error_message(e):
    """
    Print a message for a validation error.
    input: e (ValidationError): The validation error from the validate library.
    output: human readable message for the validation error.
    """
    msg = "There following error was found in your metadata:"
    e_type = e.schema_path.pop().strip()
    print(e.schema_path)
    if e_type == "type":
        s = ''
        while e.schema_path:
            p_v = e.schema_path.popleft()
            if p_v.strip() != "properties":
                if s != '':
                    s += ' -> '
                s += p_v
        msg = f"{msg} {s} needs to be a list of values."
    if e_type == "required":
        # peel out the first line from the stringified error message
        msg = f"{msg} {e.message.splitlines()[0]}"
    return msg


import requests
from .config import format_agent_profile_name
from ..constants import PENNSIEVE_URL
def create_unique_profile_name(token, machine_username_specifier):
    try:
        # get the users email

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {token}",
        }


        r = requests.get(f"{PENNSIEVE_URL}/user", headers=headers)    
        r.raise_for_status()    

        user_info = r.json()

        # create a substring of the start of the email to the @ symbol
        email = user_info["email"]
        email_sub = email.split("@")[0]

        organization_id = user_info["preferredOrganization"]

        # create an updated profile name that is unqiue to the user and their workspace 
        return format_agent_profile_name(f"soda-pennsieve-{machine_username_specifier}-{email_sub}-{organization_id.lower()}")
    except Exception as e:
        raise e


import requests
from os.path import expanduser, join
from .exceptions import FailedToFetchPennsieveDatasets, PennsieveDatasetCannotBeFound
from .authentication import get_access_token, create_request_headers


from ..constants import PENNSIEVE_URL

userpath = expanduser("~")
configpath = join(userpath, ".pennsieve", "config.ini")

def generate_options_set(soda_json_structure):
    return "generate-dataset" in soda_json_structure.keys()

def generating_locally(soda_json_structure):
    return soda_json_structure["generate-dataset"]["destination"] == "local"

def generating_on_ps(soda_json_structure):
    return soda_json_structure["generate-dataset"]["destination"] == "ps"

def uploading_with_ps_account(soda_json_structure):
    return "ps-account-selected" in soda_json_structure

def uploading_to_existing_ps_dataset(soda_json_structure):
    return "ps-dataset-selected" in soda_json_structure

def can_resume_prior_upload(resume_status):
    global ums 
    return resume_status and ums.df_mid_has_progress()

def virtual_dataset_empty(soda_json_structure):
    return (
        "dataset-structure" not in soda_json_structure
        and "metadata-files" not in soda_json_structure
        )

def get_dataset_id(dataset_name_or_id):
    """
    Returns the dataset ID for the given dataset name.
    If the dataset ID was provided instead of the name, the ID will be returned. *Common for Guided Mode*
    
    Input:
        dataset_name_or_id: Pennsieve dataset name or ID to get the ID for
    """
    # If the input is already a dataset ID, return it
    if dataset_name_or_id.startswith("N:dataset:"):
        return dataset_name_or_id
    
    # Attempt to retrieve the user's dataset list from Pennsieve
    dataset_list = get_users_dataset_list()    
    
    # Iterate through the user's dataset list to find a matching dataset name
    for dataset in dataset_list:
        if dataset["content"]["name"] == dataset_name_or_id:
            return dataset["content"]["id"]
    
    # If no matching dataset is found, abort with a 404 status and a specific error message
    raise PennsieveDatasetCannotBeFound(dataset_name_or_id)


def get_users_dataset_list():
    """
        Returns a list of datasets the user has access to.
        Input:
            token: Pennsieve access token
    """

    # The number of datasets to retrieve per chunk
    NUMBER_OF_DATASETS_PER_CHUNK = 200
    # The total number of datasets the user has access to (set after the first request)
    NUMBER_OF_DATASETS_USER_HAS_ACCESS_TO = None

    # The offset is the number of datasets to skip before retrieving the next chunk of datasets (starts at 0, then increases by the number of datasets per chunk)
    current_offset = 0
    # The list of datasets the user has access to (datasets are added to this list after each request and then returned)
    datasets = []

    try:
        # Get the first chunk of datasets as well as the total number of datasets the user has access to
        r = requests.get(f"{PENNSIEVE_URL}/datasets/paginated", headers=create_request_headers(get_access_token()), params={"offset": current_offset, "limit": NUMBER_OF_DATASETS_PER_CHUNK})
        r.raise_for_status()
        responseJSON = r.json()
        datasets.extend(responseJSON["datasets"])
        NUMBER_OF_DATASETS_USER_HAS_ACCESS_TO = responseJSON["totalCount"]

        # If the number of datasets the user has access to is less than the number of datasets per chunk, we don't need to retrieve any more datasets
        if NUMBER_OF_DATASETS_USER_HAS_ACCESS_TO < NUMBER_OF_DATASETS_PER_CHUNK:
            return datasets
        
        # Otherwise, we need to retrieve the rest of the datasets.
        # We do this by retrieving chunks of datasets until the number of datasets retrieved is equal to the number of datasets the user has access to
        while len(datasets) < NUMBER_OF_DATASETS_USER_HAS_ACCESS_TO:
            # Increase the offset by the number of datasets per chunk (e.g. if 200 datasets per chunk, then increase the offset by 200)
            current_offset += NUMBER_OF_DATASETS_PER_CHUNK
            r = requests.get(f"{PENNSIEVE_URL}/datasets/paginated", headers=create_request_headers(get_access_token()), params={"offset": current_offset, "limit": NUMBER_OF_DATASETS_PER_CHUNK})
            r.raise_for_status()
            responseJSON = r.json()
            datasets.extend(responseJSON["datasets"])

        return datasets
    except Exception as e:
        raise FailedToFetchPennsieveDatasets("Error: Failed to retrieve datasets from Pennsieve. Please try again later.")









from os.path import join, getsize
from openpyxl import load_workbook
from openpyxl.styles import Font
import shutil
import tempfile
from .helpers import upload_metadata_file, get_template_path


from .constants import METADATA_UPLOAD_PS_PATH, TEMPLATE_PATH
from .excel_utils import rename_headers, excel_columns
from ...utils import validate_schema


### Create submission file
def create_excel(soda, upload_boolean, local_destination):
    """
    Create an Excel file for submission metadata.

    Args:
        soda (dict): The soda object containing dataset metadata.
        upload_boolean (bool): Whether to upload the file to Pennsieve.
        destination_path (str): The path to save the Excel file.

    Returns:
        dict: A dictionary containing the size of the metadata file.
    """

    validate_schema(soda["dataset_metadata"]["submission"], "submission_schema.json")

    font_submission = Font(name="Calibri", size=14, bold=False)

    source = get_template_path("submission.xlsx")

    destination = join(METADATA_UPLOAD_PS_PATH, "submission.xlsx") if upload_boolean else local_destination

    try:
        shutil.copyfile(source, destination)
    except FileNotFoundError as e:
        raise e
    
    #TODO: Do not use an array for the non-array values; zipping for the sake of the ascii value is not necessary until milestone_achieved
    submission_metadata_list = [
        soda["dataset_metadata"]["submission"]
    ]

    # write to excel file
    wb = load_workbook(destination)
    ws1 = wb["Sheet1"]
    start_index = 2
    for column, submission_data in zip(excel_columns(start_index), submission_metadata_list):
        ws1[column + "2"] = submission_data["consortium_data_standard"]
        ws1[column + "3"] = submission_data["funding_consortium"]
        ws1[column + "4"] = submission_data["award_number"]
        for col, milestone in zip(excel_columns(start_index), submission_data["milestone_achieved"]):
            ws1[col + str(5)] = milestone
        ws1[column + "6"] = submission_data["milestone_completion_date"]
        ws1[column + "2"].font = font_submission
        ws1[column + "3"].font = font_submission
        ws1[column + "4"].font = font_submission
        ws1[column + "5"].font = font_submission
        ws1[column + "6"].font = font_submission

    # TODO: should milestone completion date also be an array?
    rename_headers(ws1, len(submission_metadata_list[0]["milestone_achieved"]), 2)

    wb.save(destination)

    print("Excel file created successfully at:", destination)

    wb.close()

    # calculate the size of the metadata file
    size = getsize(destination)

    

    ## if generating directly on Pennsieve, then call upload function and then delete the destination path
    if upload_boolean:
        print("Uploading Excel file to Pennsieve...")
        upload_metadata_file("submission.xlsx", soda, destination, True)
        print("Excel file uploaded successfully to Pennsieve.")
    return {"size": size}












from .constants import METADATA_UPLOAD_PS_PATH, TEMPLATE_PATH
from os.path import join, getsize
from .helpers import upload_metadata_file, get_template_path
import shutil


# this function saves and uploads the README/CHANGES to Pennsieve, just when users choose to generate onto Pennsieve
## (not used for generating locally)
def create_text_file(soda, upload_boolean, local_destination, file_type):
    """
    Create a text file for README or CHANGES metadata using a template.

    Args:
        soda (dict): The soda object containing dataset metadata.
        upload_boolean (bool): Whether to upload the file to Pennsieve.
        local_destination (str): The path to save the text file.
        file_type (str): The type of the file to be created, either "README" or "CHANGES".

    Returns:
        dict: A dictionary containing the size of the metadata file.
    """

    template_filename = f"{file_type}.txt"
    source = get_template_path(template_filename)
    destination = join(METADATA_UPLOAD_PS_PATH, template_filename) if upload_boolean else local_destination

    # Copy the template to the destination (if it exists)
    try:
        shutil.copyfile(source, destination)
    except FileNotFoundError:
        # If template not found, just create a new file
        with open(destination, "w", encoding="utf-8") as f:
            pass

    # Write the actual content from soda into the file (overwriting template content)
    text = soda["dataset_metadata"].get(file_type, "")
    with open(destination, "w", encoding="utf-8") as file:
        file.write(text)

    size = getsize(destination)
    if upload_boolean:
        upload_metadata_file(template_filename, soda, destination, True)

    return size





import requests
from ...constants import PENNSIEVE_URL
from ...utils import PennsieveActionNoPermission

def pennsieve_get_current_user_permissions(dataset_id, ps_or_token):

    if type(ps_or_token) is str:
        access_token = ps_or_token
    else:
        access_token = ps_or_token.get_user().session_token

    r = requests.get(f"{PENNSIEVE_URL}/datasets/{dataset_id}/role", headers={"Authorization": f"Bearer {access_token}"})
    r.raise_for_status()

    return r.json()


def has_edit_permissions(ps_or_token, selected_dataset_id):
    """
        Checks if the current user has permission to edit the given dataset.

        Input:
            selected_dataset_id: Pennsieve dataset ID to check permissions for
            ps: Pennsieve client object of a user that has been authenticated
    """
    try:
        role = pennsieve_get_current_user_permissions(selected_dataset_id, ps_or_token)["role"]
    except Exception as e:
        raise PennsieveActionNoPermission("edit on " + selected_dataset_id) from e

    return role in ["owner", "manager", "editor"]  


from .constants import METADATA_UPLOAD_PS_PATH, TEMPLATE_PATH, SDS_FILE_SUBJECTS,SCHEMA_NAME_SUBJECTS
from .excel_utils import rename_headers, excel_columns
from openpyxl.styles import PatternFill
from os.path import join, getsize
from openpyxl import load_workbook
import shutil
import numpy as np
from ...utils import validate_schema, get_sds_headers
from openpyxl.styles import Font
from .helpers import transposeMatrix, getMetadataCustomFields, sortedSubjectsTableData, upload_metadata_file, get_template_path


def create_excel(soda, upload_boolean, local_destination):
    source = get_template_path(SDS_FILE_SUBJECTS)


    subjects = soda["dataset_metadata"]["subjects"]

    validate_schema(subjects, SCHEMA_NAME_SUBJECTS)

    destination = join(METADATA_UPLOAD_PS_PATH, SDS_FILE_SUBJECTS) if upload_boolean else local_destination
    shutil.copyfile(source, destination)
    wb = load_workbook(destination)
    ws1 = wb["Sheet1"]



    # 1. see if the length of datastructure[0] == length of datastructure. If yes, go ahead. If no, add new columns from headers[n-1] onward.
    orangeFill = PatternFill(
        start_color="FFD965", end_color="FFD965", fill_type="solid"
    )


    # 1.1 (optional) add custom fields to the headers of the workbook
    
    # for column, header in zip(
    #     excel_columns(start_index=29), subjects[0].keys()
    # ):
    #     cell = column + str(1)
    #     ws1[cell] = header
    #     ws1[cell].fill = orangeFill
    #     ws1[cell].font = Font(bold=True, size=12, name="Calibri")

    row = 2
    ascii_headers = excel_columns(start_index=0)
    custom_headers_to_column = {}
    sds_headers = get_sds_headers(SCHEMA_NAME_SUBJECTS)

    # 2. populate excel file with the data 
    for subject in subjects:
        ws1[ascii_headers[0] + str(row)] = subject.get("subject_id", "")
        ws1[ascii_headers[0] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[1] + str(row)] = subject.get("pool_id", "")
        ws1[ascii_headers[1] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[2] + str(row)] = subject.get("subject_experimental_group", "")
        ws1[ascii_headers[2] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[3] + str(row)] = subject.get("age", "")
        ws1[ascii_headers[3] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[4] + str(row)] = subject.get("sex", "")
        ws1[ascii_headers[4] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[5] + str(row)] = subject.get("species", "")
        ws1[ascii_headers[5] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[6] + str(row)] = subject.get("strain", "")
        ws1[ascii_headers[6] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[7] + str(row)] = subject.get("rrid_for_strain", "")
        ws1[ascii_headers[7] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[8] + str(row)] = subject.get("age_category", "")
        ws1[ascii_headers[8] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[9] + str(row)] = subject.get("also_in_dataset", "")
        ws1[ascii_headers[9] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[10] + str(row)] = subject.get("member_of", "")
        ws1[ascii_headers[10] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[11] + str(row)] = subject.get("metadata_only", "")
        ws1[ascii_headers[11] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[12] + str(row)] = subject.get("laboratory_internal_id", "")
        ws1[ascii_headers[12] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[13] + str(row)] = subject.get("date_of_birth", "")
        ws1[ascii_headers[13] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[14] + str(row)] = subject.get("age_range_min", "")
        ws1[ascii_headers[14] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[15] + str(row)] = subject.get("age_range_max", "")
        ws1[ascii_headers[15] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[16] + str(row)] = subject.get("body_mass", "")
        ws1[ascii_headers[16] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[17] + str(row)] = subject.get("genotype", "")
        ws1[ascii_headers[17] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[18] + str(row)] = subject.get("phenotype", "")
        ws1[ascii_headers[18] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[19] + str(row)] = subject.get("handedness", "")
        ws1[ascii_headers[19] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[20] + str(row)] = subject.get("reference_atlas", "")
        ws1[ascii_headers[20] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[21] + str(row)] = subject.get("experimental_log_file_path", "")
        ws1[ascii_headers[21] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[22] + str(row)] = subject.get("experiment_date", "")
        ws1[ascii_headers[22] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[23] + str(row)] = subject.get("disease_or_disorder", "")
        ws1[ascii_headers[23] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[24] + str(row)] = subject.get("intervention", "")
        ws1[ascii_headers[24] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[25] + str(row)] = subject.get("disease_model", "")
        ws1[ascii_headers[25] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[26] + str(row)] = subject.get("protocol_title", "")
        ws1[ascii_headers[26] + str(row)].font = Font(bold=False, size=11, name="Arial")

        ws1[ascii_headers[27] + str(row)] = subject.get("protocol_url_or_doi", "")
        ws1[ascii_headers[27] + str(row)].font = Font(bold=False, size=11, name="Arial")

        # handle custom fields
        for field_name, field in subject.items():
            if field_name in sds_headers:
                continue

            # check if the field is already in the custom_headers_to_column dictionary
            if field_name not in custom_headers_to_column:
                print(field_name)


                custom_headers_to_column[field_name] = len(custom_headers_to_column.keys()) + 1

                # create the column header in the excel file
                offset_from_final_sds_header = custom_headers_to_column[field_name]
                ws1[ascii_headers[27 + offset_from_final_sds_header] + "1"] = field_name
                ws1[ascii_headers[27 + offset_from_final_sds_header] + "1"].fill = orangeFill
                ws1[ascii_headers[27 + offset_from_final_sds_header] + "1"].font = Font(bold=True, size=12, name="Calibri")



            # add the field value to the corresponding cell in the excel file
            offset_from_final_sds_header = custom_headers_to_column[field_name]

            ws1[ascii_headers[27 + offset_from_final_sds_header] + str(row)] = field
            ws1[ascii_headers[27 + offset_from_final_sds_header] + str(row)].font = Font(bold=False, size=11, name="Arial")


        row += 1

    wb.save(destination)

    size = getsize(destination)

    ## if generating directly on Pennsieve, then call upload function and then delete the destination path
    if upload_boolean:
        upload_metadata_file(SDS_FILE_SUBJECTS, soda, destination, True)

    return {size: size}